'''Strawberry GraphQL ASGI app with a fix for async_asgi_testclient.

Strawberry: https://strawberry.rocks
'''
from strawberry.asgi import GraphQL as GraphQL_
from strawberry.subscriptions import GRAPHQL_WS_PROTOCOL


class GraphQL(GraphQL_):
    """Add a fix to the strawberry GraphQL for async_asgi_testclient"""

    async def __call__(self, scope, receive, send):
        if scope["type"] == "websocket":
            if not scope.get("subprotocols"):
                # strawberry closes the websocket connection if
                # subprotocols are empty, which is the case for
                # async_asgi_testclient.
                scope["subprotocols"] = [GRAPHQL_WS_PROTOCOL]
        await super().__call__(scope, receive, send)



'''Configuration by Dynaconf.

Dynaconf: https://www.dynaconf.com/ 
'''
__all__ = ['load_settings']
from collections.abc import Sequence
from itertools import chain
from pathlib import Path
from typing import Optional

from apluggy import PluginManager
from dynaconf import Dynaconf, Validator

HERE = Path(__file__).resolve().parent
DEFAULT_CONFIG_PATH = HERE / 'default.toml'

assert DEFAULT_CONFIG_PATH.is_file()

MINIMAL_PRELOAD = (str(DEFAULT_CONFIG_PATH),)
MINIMAL_VALIDATORS = ()


def load_settings(hook: PluginManager) -> Dynaconf:
    '''Return a Dynaconf settings after validation'''

    settings = _load_settings(
        preload=tuple(chain(*hook.hook.dynaconf_preload())),
        settings_files=tuple(chain(*hook.hook.dynaconf_settings_files())),
        validators=tuple(chain(*hook.hook.dynaconf_validators())),
    )

    return settings


def _load_settings(
    preload: Optional[Sequence[str]] = None,
    settings_files: Optional[Sequence[str]] = None,
    validators: Optional[Sequence[Validator]] = None,
) -> Dynaconf:
    cwd = Path.cwd()
    minimal_settings_files = (str(cwd / 'nextline-graphql.toml'),)

    preload = MINIMAL_PRELOAD + tuple(preload or ())
    settings_files = minimal_settings_files + tuple(settings_files or ())
    validators = MINIMAL_VALIDATORS + tuple(validators or ())

    # Dynaconf object:
    # Dynaconf Doc: https://www.dynaconf.com/configuration/
    # Example: https://github.com/rochacbruno/learndynaconf/blob/main/config.py
    settings = Dynaconf(
        envvar_prefix="NEXTLINE",
        preload=preload,
        settings_files=settings_files,
        validators=validators,
        merge_enabled=True,
    )

    return settings



import asyncio
from typing import Any, Optional

from async_asgi_testclient import TestClient
from nextline.utils import agen_with_wait

from nextlinegraphql.plugins.ctrl.graphql import (
    MUTATE_EXEC,
    MUTATE_RESET,
    MUTATE_SEND_PDB_COMMAND,
    QUERY_STATE,
    SUBSCRIBE_PROMPTING,
    SUBSCRIBE_STATE,
    SUBSCRIBE_TRACE_IDS,
)
from nextlinegraphql.plugins.graphql.test import gql_request, gql_subscribe


async def run_statement(client, statement: Optional[str] = None):
    variables = {"statement": statement}
    data = await gql_request(client, MUTATE_RESET, variables=variables)
    assert data["reset"]

    task_subscribe_state = asyncio.create_task(subscribe_state(client))

    data = await gql_request(client, QUERY_STATE)
    assert "initialized" == data["state"]

    await asyncio.sleep(0.01)

    task_control_execution = asyncio.create_task(control_execution(client))

    await asyncio.sleep(0.01)

    data = await gql_request(client, MUTATE_EXEC)
    assert data["exec"]

    states, *_ = await asyncio.gather(
        task_subscribe_state,
        task_control_execution,
    )
    assert ["initialized", "running", "finished"] == states

    data = await gql_request(client, QUERY_STATE)
    assert "finished" == data["state"]

    await asyncio.sleep(0.01)


async def subscribe_state(client: TestClient) -> list[str]:
    ret = []
    async for data in gql_subscribe(client, SUBSCRIBE_STATE):
        s = data["state"]
        ret.append(s)
        if s == "finished":
            break
    return ret


async def control_execution(client: TestClient):
    agen = agen_with_wait(gql_subscribe(client, SUBSCRIBE_TRACE_IDS))
    data: Any

    prev_ids = set[int]()
    async for data in agen:
        if not (ids := set(data["traceIds"])):
            break
        new_ids, prev_ids = ids - prev_ids, ids
        tasks = {
            asyncio.create_task(
                control_trace(client, id_),
            )
            for id_ in new_ids
        }
        _, pending = await agen.asend(tasks)

    await asyncio.gather(*pending)


async def control_trace(client: TestClient, trace_no: int) -> None:
    # print(f'control_trace({trace_id})')

    async for data in gql_subscribe(
        client,
        SUBSCRIBE_PROMPTING,
        variables={"traceId": trace_no},
    ):
        state = data["prompting"]
        # print(state)
        if state["prompting"]:
            await asyncio.sleep(0.001)
            command = "continue"
            variables = {
                "command": command,
                "promptNo": state["prompting"],
                "traceNo": trace_no,
            }
            data = await gql_request(
                client,
                MUTATE_SEND_PDB_COMMAND,
                variables=variables,
            )
            assert data["sendPdbCommand"]



from pathlib import Path

pwd = Path(__file__).resolve().parent

sub = pwd / 'mutations'
MUTATE_EXEC = (sub / 'Exec.gql').read_text()
MUTATE_RESET = (sub / 'Reset.gql').read_text()
MUTATE_SEND_PDB_COMMAND = (sub / 'SendPdbCommand.gql').read_text()
MUTATE_RUN_AND_CONTINUE = (sub / 'RunAndContinue.gql').read_text()

sub = pwd / 'queries'
QUERY_STATE = (sub / 'State.gql').read_text()
QUERY_EXCEPTION = (sub / 'Exception.gql').read_text()
QUERY_SOURCE = (sub / 'Source.gql').read_text()
QUERY_SOURCE_LINE = (sub / 'SourceLine.gql').read_text()

sub = pwd / 'subscriptions'
SUBSCRIBE_COUNTER = (sub / 'Counter.gql').read_text()
SUBSCRIBE_STATE = (sub / 'State.gql').read_text()
SUBSCRIBE_STDOUT = (sub / 'Stdout.gql').read_text()
SUBSCRIBE_TRACE_IDS = (sub / 'TraceIds.gql').read_text()
SUBSCRIBE_PROMPTING = (sub / 'Prompting.gql').read_text()



import asyncio


def run():
    asyncio.run(c())
    return

async def c():
    a = asyncio.create_task(d())
    await asyncio.sleep(1)
    await asyncio.sleep(1)
    await a
    return

async def d():
    await asyncio.sleep(1)
    await asyncio.sleep(1)
    return




from collections.abc import MutableMapping
from typing import Any, Optional

import strawberry
from apluggy import PluginManager, asynccontextmanager
from dynaconf import Dynaconf
from starlette.applications import Starlette
from starlette.types import ASGIApp
from strawberry.schema import BaseSchema
from strawberry.tools import merge_types

from nextlinegraphql.custom.strawberry import GraphQL
from nextlinegraphql.hook import spec

from .schema import Query


class Plugin:
    @spec.hookimpl
    def configure(self, settings: Dynaconf, hook: PluginManager) -> None:
        self._settings = settings
        self._app = _create_app(hook=hook)

    @spec.hookimpl
    def schema(self):
        return (Query, None, None)

    @spec.hookimpl(tryfirst=True)  # tryfirst so to be the outermost context
    @asynccontextmanager
    async def lifespan(self, app: Starlette):
        app.mount('/', self._app)
        yield

    @spec.hookimpl
    def update_strawberry_context(self, context: MutableMapping[str, Any]) -> None:
        context['settings'] = self._settings


def _create_app(hook: PluginManager) -> ASGIApp:
    schema = _compose_schema(hook=hook)
    app = _EGraphQL(schema).set_hook(hook)
    return app


def _compose_schema(hook: PluginManager) -> BaseSchema:
    # [(Query, Mutation, Subscription), ...]
    three_types = hook.hook.schema()

    # [(Query, ...), (Mutation, ...), (Subscription, ...)]
    transposed = list(map(tuple, zip(*three_types)))
    transposed = [tuple(t for t in l if t) for l in transposed]  # remove None

    q, m, s = transposed
    assert q  # Query is required
    Query = merge_types('Query', q)  # type: ignore
    Mutation = m and merge_types('Mutation', m) or None  # type: ignore
    Subscription = s and merge_types('Subscription', s) or None  # type: ignore

    schema = strawberry.Schema(
        query=Query,
        mutation=Mutation,
        subscription=Subscription,
    )

    return schema


class _EGraphQL(GraphQL):
    '''Extend the strawberry GraphQL app to override the `get_context` method

    This class is implemented in the way described in the strawberry document:
    https://strawberry.rocks/docs/integrations/asgi
    '''

    def set_hook(self, hook: PluginManager) -> '_EGraphQL':
        self._hook = hook
        return self

    async def get_context(self, request, response=None) -> Optional[Any]:
        context = {'request': request, 'response': response}
        self._hook.hook.update_strawberry_context(context=context)
        return context



import json
from typing import cast

import strawberry
from starlette.requests import Request
from strawberry.types import Info


def query_headers(info: Info) -> str:
    request = cast(Request, info.context['request'])
    return json.dumps(dict(request.headers))


@strawberry.type
class QueryDev:
    headers: str = strawberry.field(resolver=query_headers)


@strawberry.type
class Query:
    @strawberry.field
    def dev(self) -> QueryDev:
        return QueryDev()



from apluggy import PluginManager
from dynaconf import Dynaconf

from nextlinegraphql.hook import spec

from .schema import Query


class Plugin:
    @spec.hookimpl
    def configure(self, settings: Dynaconf, hook: PluginManager) -> None:
        self._settings = settings

    @spec.hookimpl
    def schema(self):
        return (Query, None, None)



'''Hook specification for Nextline GraphQL plugin.'''
from collections.abc import MutableMapping, Sequence
from typing import Optional

import apluggy as pluggy
from apluggy import asynccontextmanager
from dynaconf import Dynaconf, Validator
from starlette.applications import Starlette

PROJECT_NAME = 'nextline_graphql'

hookspec = pluggy.HookspecMarker(PROJECT_NAME)
hookimpl = pluggy.HookimplMarker(PROJECT_NAME)


@hookspec
def dynaconf_preload() -> Optional[Sequence[str]]:
    '''Return settings file paths of the plugin to be preloaded.

    The paths will be included in the "preload" option of Dynaconf.

    https://www.dynaconf.com/configuration/#preload
    '''


@hookspec
def dynaconf_settings_files() -> Optional[Sequence[str]]:
    '''Return settings file paths of the plugin.

    The paths will be included in the "settings_files" option of Dynaconf.

    https://www.dynaconf.com/configuration/#settings_file-or-settings_files
    '''


@hookspec
def dynaconf_validators() -> Optional[Sequence[Validator]]:
    '''Return validators of the plugin settings.

    The validators will be included in the "validators" option of Dynaconf.

    https://www.dynaconf.com/configuration/#validators
    '''


@hookspec
def configure(settings: Dynaconf, hook: pluggy.PluginManager) -> None:
    '''Initialize the plugin.'''


@hookspec
def schema() -> Optional[tuple[type, type | None, type | None]]:
    '''The GraphQL schema (Query, Mutation, Subscription)'''


@hookspec
async def update_lifespan_context(
    app: Starlette, hook: pluggy.PluginManager, context: MutableMapping
) -> None:
    '''Update the context object before it is passed to the lifespan hook.'''


@hookspec
@asynccontextmanager
async def lifespan(app: Starlette, hook: pluggy.PluginManager, context: MutableMapping):
    '''Called within the Starlette lifespan context.

    The context is passed from the update_lifespan_context hook.

    The Starlette lifespan yields within this hook
    '''


@hookspec
def update_strawberry_context(context: MutableMapping) -> None:
    '''Called within get_context() of a subclass of strawberry.asgi.GraphQL.

    Plugins can modify the context in place.

    Strawberry Doc: https://strawberry.rocks/docs/integrations/asgi#get_context
    '''



__all__ = ['Plugin']

import asyncio
from collections.abc import AsyncIterator, MutableMapping

import apluggy as pluggy
from apluggy import asynccontextmanager
from nextline import Nextline
from nextline.types import StdoutInfo
from nextline.utils import merge_aiters

from nextlinegraphql.hook import spec

from .example_script import statement
from .schema import Mutation, Query, Subscription


class Plugin:
    @spec.hookimpl
    def schema(self):
        return (Query, Mutation, Subscription)

    @spec.hookimpl
    async def update_lifespan_context(self, context: MutableMapping) -> None:
        self._nextline = Nextline(statement)
        self._stdout_cache = list[str]()
        context['nextline'] = self._nextline

    @spec.hookimpl(trylast=True)  # trylast so to be the innermost context
    @asynccontextmanager
    async def lifespan(self, hook: pluggy.PluginManager) -> AsyncIterator[None]:
        '''Yield within the nextline context.'''
        async with (self._nextline, self._cache_stdout()):
            yield

    @spec.hookimpl
    def update_strawberry_context(self, context: MutableMapping) -> None:
        context['nextline'] = self._nextline
        context['stdout_cache'] = self._stdout_cache

    @asynccontextmanager
    async def _cache_stdout(self) -> AsyncIterator[None]:
        task = asyncio.create_task(self.__cache_stdout())
        try:
            yield
        finally:
            task.cancel()
            try:
                await task
            except asyncio.CancelledError:
                pass

    async def __cache_stdout(self) -> None:
        aiters = merge_aiters(
            self._nextline.subscribe_state(), self._nextline.subscribe_stdout()
        )
        async for _, v in aiters:
            match v:
                case 'initialized':
                    self._stdout_cache.clear()
                case StdoutInfo(text=text) if text is not None:
                    self._stdout_cache.append(text)
